from importlib import import_module
from typing import Type

from shard_static import config
from shard_static.lock import BaseLockManager


def get_lock_manager(model_name: str, database_alias: str) -> BaseLockManager:
    lock_key = _make_lock_key(model_name, database_alias)
    ttl = config.SHARD_SYNC_LOCK_TTL
    lock_manager_class = _import_lock_manager_class()

    return lock_manager_class(key=lock_key, ttl=ttl)


def _make_lock_key(model_name: str, database_alias: str) -> str:
    return f'{config.SHARD_SYNC_LOCK_KEY_PREFIX}:{model_name}:{database_alias}'


def _import_lock_manager_class() -> Type[BaseLockManager]:
    mod_path, _, cls_name = config.SHARD_SYNC_LOCK_MANAGER_CLASS.rpartition('.')
    mod = import_module(mod_path)
    cls = getattr(mod, cls_name)
    return cls



from django.conf import settings

from shard.exceptions import NotShardingModelException, RequireShardKeyException
from shard.mixins import ShardMixin

__all__ = ('get_shard_key_from_kwargs', 'get_shard_key_from_instance', 'mod_shard_key_by_replica_count', )

SHARD_REPLICA_COUNT_SETTING = 'SHARD_REPLICA_COUNT_SETTING'
DEFAULT_REPLICA_COUNT = 1024


def get_shard_key_from_kwargs(model, **kwargs) -> int:
    _validate_shard_model(model)

    if model.shard_key_name not in kwargs:
        raise RequireShardKeyException()

    return kwargs[model.shard_key_name]


def get_shard_key_from_instance(instance: ShardMixin) -> int:
    return instance.get_shard_key()


def mod_shard_key_by_replica_count(shard_key: int, shard_group: str) -> int:
    return _calc_shard_key(shard_key, shard_group)


def _validate_shard_model(model):
    if not issubclass(model, ShardMixin):
        raise NotShardingModelException()

    if model.shard_group is None or model.shard_key_name is None:
        raise NotShardingModelException()


def _calc_shard_key(key: int, shard_group: str) -> int:
    return int(key % _get_replica_count(shard_group=shard_group))


def _get_replica_count(shard_group: str) -> int:
    replica_settings = _get_shard_replica_settings()
    return replica_settings.get(shard_group, DEFAULT_REPLICA_COUNT)


def _get_shard_replica_settings():
    return getattr(settings, SHARD_REPLICA_COUNT_SETTING, {})



import typing

from django.db.models.manager import Manager

from shard.queryset import ShardQuerySet
from shard.utils.shard import get_shard_by_shard_key_and_shard_group
from shard.utils.shard_key import get_shard_key_from_kwargs

__all__ = ('ShardManager', )


def _wrap(func_name):
    def wrapped(self, *args, **kwargs):
        shard_key = get_shard_key_from_kwargs(model=self.model, **kwargs)
        return getattr(self.get_queryset(shard_key=shard_key), func_name)(*args, **kwargs)

    wrapped.__name__ = func_name
    return wrapped


class BaseShardManager(Manager):
    def shard(self, shard: str):
        return self.get_queryset().using(shard)

    def get_queryset(self, shard_key=None):
        hints = {'shard_key': shard_key}
        return ShardQuerySet(model=self.model, hints=hints)


class ShardManager(BaseShardManager):
    def raw(self, shard_key: int, query: str):
        return self.shard(shard=get_shard_by_shard_key_and_shard_group(shard_key=shard_key, shard_group=self.model.shard_group)).raw(query)

    def bulk_create(self, shard_key: int, objs: typing.List, batch_size: int):
        return self.shard(shard=get_shard_by_shard_key_and_shard_group(shard_key=shard_key, shard_group=self.model.shard_group))\
            .bulk_create(objs=objs, batch_size=batch_size)

    filter = _wrap('filter')
    get = _wrap('get')
    create = _wrap('create')
    get_or_create = _wrap('get_or_create')
    update_or_create = _wrap('update_or_create')



from typing import Dict

from shard.config.database_config import make_shard_configuration, make_replication_configuration

__all__ = ('ConfigHelper', )


class ConfigHelper:
    @staticmethod
    def database_configs(unshard: Dict=None, shard: Dict=None) -> Dict:
        configuration = {}

        # Make Unshard Configuration
        if unshard:
            for key, config in unshard.items():
                configuration.update(make_replication_configuration(key=key, replication_config=config))

        # Make Shard Configuration
        if shard:
            for shard_group, config in shard.items():
                configuration.update(make_shard_configuration(
                    shard_group=shard_group, shard_options=config['shard_options'], shards=config['shards'],
                    db_options=config.get('db_options')
                ))

        return configuration



import logging
from datetime import datetime
from typing import Optional, Type, List

from django.apps import apps
from django.conf import settings
from django.db import transaction
from django.forms import model_to_dict

from shard.constants import DATABASE_CONFIG_SHARD_GROUP, DEFAULT_DATABASE
from shard.utils.database import get_master_databases_by_shard_group
from shard_static import config
from shard_static.exceptions import InvalidDatabaseAliasException, NotDiffusibleException, NotShardStaticException
from shard_static.models import BaseShardStaticModel, StaticSyncStatus
from shard_static.services import lock_service

logger = logging.getLogger('shard_static.services.sync_static_service')


def run_sync_with_lock(model_name: str, database_alias: str):
    lock_manager = lock_service.get_lock_manager(model_name, database_alias)
    if not lock_manager.lock():
        logger.info(f'Already exists processing - {model_name}, {database_alias}')
        return

    try:
        run_sync(model_name=model_name, database_alias=database_alias)
    finally:
        lock_manager.release()


def run_sync(model_name: str, database_alias: str):
    model = _get_model(model_name=model_name)

    _validate_model(model)
    _validate_database(model, database_alias)

    sync_status, _ = StaticSyncStatus.objects.shard(shard=database_alias).get_or_create(static_model_key=_make_sync_key(model=model))

    offset = 0
    limit = config.SHARD_SYNC_MAX_ITEMS
    while True:
        source_items = model.objects.find_by_last_modified(last_modified=sync_status.last_modified, offset=offset, limit=limit)
        logger.debug(
            f'[Load source items] - last_modified: {sync_status.last_modified}, offset: {offset}, limit: {limit}',
            extra={'offset': offset, 'limit': limit, 'last_modified':sync_status.last_modified}
        )

        last_modified = _insert_items(items=source_items, model=model, database_alias=database_alias)

        if sync_status.last_modified != last_modified:
            sync_status.last_modified = last_modified
            sync_status.save(using=database_alias)
            return

        offset += limit


def _insert_items(items: List, model: Type[BaseShardStaticModel], database_alias) -> datetime:
    last_modified = None
    with transaction.atomic(database_alias):
        for _data in items:
            if not last_modified or last_modified < _data.last_modified:
                last_modified = _data.last_modified

            model.objects.shard(shard=database_alias).update_or_create(
                id=_data.id,
                defaults=model_to_dict(
                    _data, fields=[field.name for field in _data._meta.fields]  # flake8: noqa: W0212 pylint: disable=protected-access
                )
            )

    return last_modified


def _get_model(model_name: str) -> Type[BaseShardStaticModel]:
    _splited = model_name.split('.')
    _label, _model_name = _splited[0], _splited[1]

    app = apps.get_app_config(_label)
    model = app.get_model(_model_name)
    return model


def _validate_model(model: Type[BaseShardStaticModel]):
    if not issubclass(model, BaseShardStaticModel):
        raise NotShardStaticException()

    if not model.diffusible:
        raise NotDiffusibleException()


def _validate_database(model: Type[BaseShardStaticModel], database_alias: str):
    if database_alias == DEFAULT_DATABASE:
        raise InvalidDatabaseAliasException()

    if _get_shard_group_from_database(database_alias) is None:
        raise InvalidDatabaseAliasException()

    databases = get_master_databases_by_shard_group(shard_group=model.shard_group)
    if database_alias not in databases:
        raise InvalidDatabaseAliasException()


def _get_shard_group_from_database(database_alias: str) -> Optional[str]:
    db_setting = settings.DATABASES.get(database_alias, None)
    if db_setting:
        return db_setting.get(DATABASE_CONFIG_SHARD_GROUP, None)

    return None


def _make_sync_key(model: Type[BaseShardStaticModel]):
    return model._meta.db_table  # flake8: noqa: W0212 pylint: disable=protected-access



from typing import List, Dict

from django.conf import settings

from shard.utils.memorize import memorize
from shard.constants import DATABASE_CONFIG_MASTER, DATABASE_CONFIG_SHARD_GROUP, DATABASE_CONFIG_SHARD_NUMBER

__all__ = (
    'get_master_databases', 'get_master_databases_for_shard', 'get_master_databases_by_shard_group', 'get_slave_databases_by_master',
)


@memorize
def get_master_databases(without_shard: bool = True) -> List:
    databases = _get_databases()
    result = []

    for key, config in databases.items():
        if config.get(DATABASE_CONFIG_MASTER):
            continue

        if without_shard:
            if config.get(DATABASE_CONFIG_SHARD_GROUP) or config.get(DATABASE_CONFIG_SHARD_NUMBER):
                continue

        result.append(key)

    return result


@memorize
def get_master_databases_for_shard() -> List:
    databases = _get_databases()
    shards = []

    for key, config in databases.items():
        if config.get(DATABASE_CONFIG_MASTER):
            continue

        if config.get(DATABASE_CONFIG_SHARD_GROUP) is None or config.get(DATABASE_CONFIG_SHARD_NUMBER) is None:
            continue

        shards.append(key)

    shards = sorted(
        shards, key=lambda shard: (databases[shard][DATABASE_CONFIG_SHARD_GROUP], databases[shard][DATABASE_CONFIG_SHARD_NUMBER])
    )
    return shards


@memorize
def get_master_databases_by_shard_group(shard_group: str) -> List:
    databases = _get_databases()
    shards = []

    for key, config in databases.items():
        if config.get(DATABASE_CONFIG_MASTER):
            continue

        _group = config.get(DATABASE_CONFIG_SHARD_GROUP, None)
        if _group is None or _group != shard_group:
            continue

        shards.append(key)

    shards = sorted(shards, key=lambda shard: databases[shard][DATABASE_CONFIG_SHARD_NUMBER])
    return shards


@memorize
def get_slave_databases_by_master(master: str) -> List:
    result = []
    for key, config in _get_databases().items():
        _master = config.get(DATABASE_CONFIG_MASTER, None)
        if _master is None or _master != master:
            continue

        result.append(key)

    return result


def _get_databases() -> Dict:
    return settings.DATABASES



from django.apps import apps

from shard.mixins import SpecificDatabaseMixin
from shard.routers.base import BaseRouter


class SpecificRouter(BaseRouter):
    def db_for_read(self, model, **hints):
        return self._get_specific_database(model=model, **hints)

    def db_for_write(self, model, **hints):
        return self._get_specific_database(model=model, **hints)

    def allow_relation(self, obj1, obj2, **hints):
        obj1_db = self._get_specific_database(model=obj1.__class__)
        obj2_db = self._get_specific_database(model=obj2.__class__)

        if obj1_db is None or obj2_db is None:
            return None

        return obj1_db == obj2_db

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        model_name = model_name or hints.get('model_name')
        model = hints.get('model')
        if model:
            model_name = model.__name__

        if "." in model_name:
            _app_label = model_name.split('.')[0]
            app = apps.get_app_config(_app_label)
            model = app.get_model(model_name[len(_app_label) + 1:])
        else:
            app = apps.get_app_config(app_label)
            model = app.get_model(model_name)

        specific_db = self._get_specific_database(model)
        if specific_db is None:
            return None

        return specific_db == db

    def _get_specific_database(self, model, **hints):
        if not issubclass(model, SpecificDatabaseMixin):
            return None

        return model.specific_database()



from django.db import connections, transaction

from shard.exceptions import QueryExecuteFailureException
from shard.utils.database import get_master_databases_by_shard_group


class QueryExecutor:
    """
    해당 클래스는 DDL 같은 Master 샤드들에 일괄적으로 실행될 필요가 있는 쿼리들을 실행하는 유틸입니다.
    """
    def __init__(self, shard_group: str):
        self._shard_group = shard_group
        self._shards = get_master_databases_by_shard_group(shard_group=shard_group)

    def run_query(self, query: str):
        executed = []
        try:
            for shard in self._shards:
                with transaction.atomic(shard):
                    cursor = connections[shard].cursor()
                    cursor.execute(query)
                executed.append(shard)
        except Exception as e:
            raise QueryExecuteFailureException(shard_group=self._shard_group, executed=executed, exception=e)

        return executed

    def get_shards(self):
        return self._shards



# Generated by Django 2.0.2 on 2018-04-11 10:26

from django.db import migrations, models
import shard_static.mixins


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='StaticSyncStatus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('static_model_key', models.CharField(max_length=64, unique=True, verbose_name='Static Model Key')),
                ('last_modified', models.DateTimeField(null=True, verbose_name='Last Modified')),
            ],
            options={
                'db_table': 'static_sync_status',
            },
            bases=(shard_static.mixins.ShardStaticMixin, models.Model),
        ),
    ]



# Generated by Django 2.0.2 on 2018-05-15 07:49

from django.db import migrations, models
import shard_static.models


class Migration(migrations.Migration):

    dependencies = [
        ('shard_static', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='staticsyncstatus',
            name='last_modified',
            field=models.DateTimeField(default=shard_static.models.get_first_day, verbose_name='Last Modified'),
        ),
    ]



from typing import List

__all__ = (
    'StrategyNotImplementException', 'RequireSpecificDatabaseException', 'RequireMasterConfigException', 'NotShardingModelException',
    'RequireShardKeyException', 'QueryExecuteFailureException',
)


class StrategyNotImplementException(Exception):
    pass


class RequireSpecificDatabaseException(Exception):
    pass


class RequireMasterConfigException(Exception):
    pass


class NotShardingModelException(Exception):
    pass


class RequireShardKeyException(Exception):
    pass


class QueryExecuteFailureException(Exception):
    def __init__(self, shard_group: str, executed: List, exception: Exception, *args, **kwargs):
        self.shard_group = shard_group
        self.executed = executed
        self.parent_exception = exception

        super().__init__(*args, **kwargs)



from django.apps import apps
from django.db import connections, transaction
from django.utils.deconstruct import deconstructible

from shard.models import TableStrategyModel
from shard.strategy.id_generation.base import BaseIDGenerationStrategy

__all__ = ('TableStrategy', )


@deconstructible
class TableStrategy(BaseIDGenerationStrategy):
    def __init__(self, source_model):
        self.source_model = source_model

    def get_next_id(self, instance):
        model = self._get_model()

        if not issubclass(model, TableStrategyModel):
            raise ValueError('ID발급에 사용할 수 없는 모델입니다.')

        model_db = getattr(model, '_specific_database', 'default')
        with transaction.atomic(model_db):
            cursor = connections[model_db].cursor()
            sql = 'INSERT INTO %s () VALUES ()' % (
                model._meta.db_table,  # flake8: noqa: W0212  # pylint: disable=protected-access
            )

            cursor.execute(sql)

        return cursor.cursor.lastrowid

    def _get_model(self):
        splited = self.source_model.split('.')

        app_label = splited[0]
        app = apps.get_app_config(app_label)
        return app.get_model(splited[-1])



from django.conf import settings

from shard_static.constants import DEFAULT_MAX_ITEMS, DEFAULT_LOCK_TTL

SHARD_SYNC_MAX_ITEMS = getattr(settings, 'SHARD_SYNC_MAX_ITEMS', DEFAULT_MAX_ITEMS)

SHARD_SYNC_LOCK_KEY_PREFIX = getattr(settings, 'SHARD_SYNC_LOCK_KEY_PREFIX', 'shard_sync_lock')
SHARD_SYNC_LOCK_TTL = getattr(settings, 'SHARD_SYNC_LOCK_TTL', DEFAULT_LOCK_TTL)
SHARD_SYNC_LOCK_MANAGER_CLASS = getattr(settings, 'SHARD_SYNC_LOCK_MANAGER_CLASS')



from datetime import datetime, timezone

from django.db import models

from shard_static.managers import ShardStaticManager, ShardStaticStatusManager
from shard_static.mixins import ShardStaticMixin


def get_first_day() -> datetime:
    return datetime(1970, 1, 1, tzinfo=timezone.utc)


# Don't link StaticSyncStatus and other model.
# StaticSyncStatus is isolation each shards.
class StaticSyncStatus(models.Model):
    static_model_key = models.CharField(max_length=64, unique=True, verbose_name='Static Model Key')
    last_modified = models.DateTimeField(null=False, default=get_first_day, verbose_name='Last Modified', )

    objects = ShardStaticStatusManager()

    class Meta:
        db_table = 'static_sync_status'


class BaseShardStaticModel(ShardStaticMixin, models.Model):
    last_modified = models.DateTimeField(auto_now=True, db_index=True, verbose_name='Last Modified', )

    objects = ShardStaticManager()

    class Meta:
        abstract = True



from datetime import datetime
from typing import Optional

from shard.managers import BaseShardManager
from shard_static.exceptions import NotExistsOriginalDataException, DontExecuteException


def _wrap_for_static(func_name):
    def wrapped(self, *args, **kwargs):
        if not self.model.diffusible:
            raise NotExistsOriginalDataException()

        return getattr(self.get_queryset(), func_name)(*args, **kwargs)

    wrapped.__name__ = func_name
    return wrapped


def _wrap_for_status(func_name):
    def wrapped(self, *args, **kwargs):
        raise DontExecuteException()

    wrapped.__name__ = func_name
    return wrapped


class ShardStaticStatusManager(BaseShardManager):
    filter = _wrap_for_status('filter')
    get = _wrap_for_status('get')
    create = _wrap_for_status('create')
    get_or_create = _wrap_for_status('get_or_create')
    update_or_create = _wrap_for_status('update_or_create')


class ShardStaticManager(BaseShardManager):
    filter = _wrap_for_static('filter')
    get = _wrap_for_static('get')
    create = _wrap_for_static('create')
    get_or_create = _wrap_for_static('get_or_create')
    update_or_create = _wrap_for_static('update_or_create')

    def find_by_last_modified(self, last_modified: Optional[datetime], offset: int, limit: int):
        qs = self.get_queryset()

        if last_modified:
            qs = qs.filter(last_modified__gt=last_modified)

        return qs[offset:offset + limit]



from typing import Callable


def memorize(func: Callable):
    _cache = {}

    def _decorator(*args, clear: bool=False, **kwargs):
        key = func.__name__ + str(args) + str(kwargs)
        if clear or key not in _cache:
            _cache[key] = func(*args, **kwargs)
        return _cache[key]

    return _decorator



from typing import List, Dict

from shard.utils.database import get_master_databases_for_shard
from shard_static.models import StaticSyncStatus


def get_sync_status() -> Dict:
    databases = get_master_databases_for_shard()

    raw_datas = []
    for database in databases:
        raw_datas.append((database, StaticSyncStatus.objects.shard(shard=database).all()))

    return _assemble_status_data(raw_datas=raw_datas)


# [qs, qs, qs] -> {shard_model_key_one: {shard_1: last_modified}, shard_model_key_two: {}, }
def _assemble_status_data(raw_datas: List) -> Dict:
    result = {}

    for database, raw_data in raw_datas:
        for obj in raw_data:
            if obj.static_model_key not in result:
                result[obj.static_model_key] = {}
            result[obj.static_model_key][database] = obj.last_modified

    return result





class NotShardStaticException(Exception):
    pass


class NotDiffusibleException(Exception):
    pass


class InvalidDatabaseAliasException(Exception):
    pass


class NotExistsOriginalDataException(Exception):
    pass


class DontExecuteException(Exception):
    pass


class DontLinkException(Exception):
    pass



from django.db import models

from shard.mixins import StrategyPkMixin


class TableStrategyPkField(StrategyPkMixin, models.BigAutoField):
    def __init__(self, source_model: str, *args, **kwargs):
        from shard.strategy.id_generation.table import TableStrategy
        self.source_model = source_model
        kwargs['strategy'] = TableStrategy(source_model=self.source_model)
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs['source_model'] = self.source_model

        return name, path, args, kwargs


class UUID4StrategyPkField(StrategyPkMixin, models.CharField):
    def __init__(self, *args, **kwargs):
        from shard.strategy.id_generation.uuid import UUIDStrategy
        kwargs['strategy'] = UUIDStrategy()

        # If null is false, field's default value is empty string.
        # So, model not call get_pk_value_on_save
        # If you will use this field, give up using migrations.
        # Crash django when using migrations, because django not supported nullable pk.
        kwargs['null'] = True
        super().__init__(*args, **kwargs)


class NanosecondStrategyPkField(StrategyPkMixin, models.BigAutoField):
    def __init__(self, *args, **kwargs):
        from shard.strategy.id_generation.nanosecond import NanosecondStrategy
        kwargs['strategy'] = NanosecondStrategy()
        super().__init__(*args, **kwargs)



from django.shortcuts import render
from django.views import View

from shard_static.services import sync_status_monitoring_service


class MonitoringView(View):
    def get(self, request):
        context = {
            'data': sync_status_monitoring_service.get_sync_status()
        }
        return render(request, 'monitoring.html', context)



# Generated by Django 2.0.2 on 2018-05-02 11:19

from django.db import migrations, models
import shard.fields
import shard.mixins
import shard.strategy.id_generation.table
import shard_static.mixins


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='IdGenerateTestModel',
            fields=[
                ('id', shard.fields.TableStrategyPkField(primary_key=True, serialize=False, source_model='tests.TestIds', strategy=shard.strategy.id_generation.table.TableStrategy(source_model='tests.TestIds'))),
            ],
        ),
        migrations.CreateModel(
            name='NormalModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
        ),
        migrations.CreateModel(
            name='ShardModelA',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('user_id', models.IntegerField(verbose_name='유저 idx')),
                ('text', models.CharField(blank=True, max_length=64, verbose_name='더미 텍스트')),
            ],
            bases=(shard.mixins.ShardMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ShardModelB',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('user_id', models.IntegerField(verbose_name='유저 idx')),
            ],
            bases=(shard.mixins.ShardMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ShardStaticA',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_modified', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Last Modified')),
            ],
            options={
                'abstract': False,
            },
            bases=(shard_static.mixins.ShardStaticMixin, models.Model),
        ),
        migrations.CreateModel(
            name='ShardStaticB',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_modified', models.DateTimeField(auto_now=True, db_index=True, verbose_name='Last Modified')),
            ],
            options={
                'abstract': False,
            },
            bases=(shard_static.mixins.ShardStaticMixin, models.Model),
        ),
        migrations.CreateModel(
            name='TestIds',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
            ],
            options={
                'abstract': False,
            },
            bases=(shard.mixins.SpecificDatabaseMixin, models.Model),
        ),
    ]



from typing import List

from shard.utils.database import get_slave_databases_by_master


class BaseReadRoutingStrategy:
    @classmethod
    def get_slave_databases(cls, master: str) -> List:
        return get_slave_databases_by_master(master=master)

    @classmethod
    def pick_read_db(cls, master: str) -> str:
        raise NotImplementedError



# pylint: disable=too-many-return-statements
from typing import Optional

from django.apps import apps
from django.conf import settings

from shard.constants import DATABASE_CONFIG_SHARD_GROUP, DEFAULT_DATABASE
from shard.mixins import ShardMixin
from shard.routers.base import BaseReplicationRouter
from shard.utils.shard import get_shard_by_shard_key_and_shard_group, get_shard_by_instance
from shard_static.exceptions import DontLinkException
from shard_static.mixins import ShardStaticMixin
from shard_static.models import StaticSyncStatus


class ShardStaticRouter(BaseReplicationRouter):
    def allow_relation(self, obj1, obj2, **hints):
        super_allow_relation = super().allow_relation(obj1, obj2, **hints)
        if super_allow_relation is not None:
            return super_allow_relation

        # Don't link StaticSyncStatus and other model.
        # StaticSyncStatus is isolation each shards.
        if isinstance(obj1, StaticSyncStatus) or isinstance(obj2, StaticSyncStatus):
            return False

        if isinstance(obj1, (ShardMixin, ShardStaticMixin)) and isinstance(obj2, (ShardMixin, ShardStaticMixin)):
            # obj1 and obj2 are for shard
            return obj1.shard_group == obj2.shard_group

        if (isinstance(obj1, (ShardMixin, ShardStaticMixin)) and not isinstance(obj2, (ShardMixin, ShardStaticMixin))) or \
                (not isinstance(obj1, (ShardMixin, ShardStaticMixin)) and isinstance(obj2, (ShardMixin, ShardStaticMixin))):

            # If you make relation shard obj and normal obj, raise exception.
            raise DontLinkException()

        # obj1 and obj2 are normal object.
        return None

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        super_allow_migrate = super().allow_migrate(db, app_label, model_name, **hints)
        if super_allow_migrate is not None:
            return super_allow_migrate

        model_name = model_name or hints.get('model_name')
        model = hints.get('model')
        if model:
            model_name = model.__name__

        if "." in model_name:
            _app_label = model_name.split('.')[0]
            app = apps.get_app_config(_app_label)
            model = app.get_model(model_name[len(_app_label) + 1:])
        else:
            app = apps.get_app_config(app_label)
            model = app.get_model(model_name)

        shard_group_for_db = settings.DATABASES[db].get(DATABASE_CONFIG_SHARD_GROUP, None)
        if not issubclass(model, (ShardMixin, ShardStaticMixin, StaticSyncStatus)):
            if shard_group_for_db:
                return False
            return None

        if issubclass(model, StaticSyncStatus):
            if db == DEFAULT_DATABASE or shard_group_for_db is None:
                return False
            return True

        if issubclass(model, ShardStaticMixin):
            if model.diffusible and db == DEFAULT_DATABASE:
                return True

        return shard_group_for_db == model.shard_group

    def _get_master_database(self, model, **hints) -> Optional[str]:
        if not issubclass(model, ShardMixin):
            return None

        shard = None
        if hints.get('shard_key'):
            shard = get_shard_by_shard_key_and_shard_group(shard_key=hints['shard_key'], shard_group=model.shard_group)
        elif hints.get('instance'):
            shard = get_shard_by_instance(instance=hints['instance'])

        return shard



import urllib.parse
from typing import Optional, Tuple, Dict, List

from dj_database_url import config

from shard.constants import DATABASE_CONFIG_MASTER, DATABASE_CONFIG_SHARD_GROUP, DATABASE_CONFIG_SHARD_NUMBER
from shard.exceptions import RequireMasterConfigException

__all__ = ('make_shard_configuration', 'make_replication_configuration', )


def make_shard_configuration(shard_group: str, shard_options: Dict, shards: List, db_options: Optional[Dict]=None) -> Dict:
    configuration = {}
    database_name = shard_options['database_name']
    shard_count = len(shards)
    logical_count_per_shard = _get_logical_count_per_shard(logical_count=shard_options['logical_count'], shard_count=shard_count)

    for index, shard_config in enumerate(shards):
        start, end = _get_logical_range(index, logical_count_per_shard)
        for logical_index in range(start, end):
            shard_name = _make_shard_name(shard_group=shard_group, shard_count=shard_count, shard_index=index, logical_index=logical_index)
            replication_config = {
                'master': _make_shard_database_url(
                    origin=shard_config['master'], database_name=database_name, logical_index=logical_index, db_options=db_options
                ),
                'slaves': [
                    _make_shard_database_url(
                        origin=slave_database_url, database_name=database_name, logical_index=logical_index, db_options=db_options
                    )
                    for slave_database_url in shard_config.get('slaves', [])
                ],
            }

            configuration.update(make_replication_configuration(key=shard_name, replication_config=replication_config, options={
                'shard_group': shard_group, 'shard_number': logical_index
            }))

    return configuration


def make_replication_configuration(key: str, replication_config: Dict, options: Dict=None) -> Dict:
    if replication_config.get('master', None) is None:
        raise RequireMasterConfigException()

    configuration = {}
    configuration[key] = _generate_database_config(
        database_url=replication_config['master'], **(options or {})
    )

    slaves = replication_config.get('slaves', [])
    for index, database_url in enumerate(slaves):
        slave_key = '%s_slave_%d' % (key, index)
        configuration[slave_key] = _generate_database_config(database_url=database_url, is_replica_of=key)

    return configuration


def _generate_database_config(
        database_url: str, is_replica_of: Optional[str]=None, shard_group: Optional[str]=None, shard_number: Optional[int]=None
) -> Dict:
    db_config = config(default=database_url)

    if is_replica_of:
        db_config[DATABASE_CONFIG_MASTER] = is_replica_of

    if shard_group is not None and shard_number is not None:
        db_config[DATABASE_CONFIG_SHARD_GROUP] = shard_group
        db_config[DATABASE_CONFIG_SHARD_NUMBER] = shard_number

    return db_config


def _get_logical_count_per_shard(logical_count: int, shard_count: int) -> int:
    return int(logical_count / shard_count)


def _get_logical_range(shard_index: int, logical_count_per_shard: int) -> Tuple[int, int]:
    start = shard_index * logical_count_per_shard
    end = (shard_index + 1) * logical_count_per_shard

    return start, end


def _make_shard_database_url(origin: str, database_name: str, logical_index: int, db_options: Optional[Dict]=None) -> str:
    if origin == 'sqlite://:memory:':
        return origin

    database_name = '%s_%d' % (database_name, logical_index)
    database_url = '%s%s' % (origin, database_name)

    if db_options:
        database_url = database_url + '?' + urllib.parse.urlencode(db_options)

    return database_url


def _make_shard_name(shard_group: str, shard_count: int, shard_index: int, logical_index: int) -> str:
    return '%s_%d_%d_%d' % (shard_group, shard_count, shard_index, logical_index)




from shard.mixins import ShardMixin
from shard.utils.shard_key import get_shard_key_from_instance, mod_shard_key_by_replica_count
from shard.utils.database import get_master_databases_by_shard_group

__all__ = (
    'get_shard_by_shard_key_and_shard_group', 'get_shard_by_instance',
)


def get_shard_by_shard_key_and_shard_group(shard_key: int, shard_group: str) -> str:
    key = mod_shard_key_by_replica_count(shard_key=shard_key, shard_group=shard_group)
    databases = get_master_databases_by_shard_group(shard_group=shard_group)
    return databases[_calc_shard(shard_key=key, shard_count=len(databases))]


def get_shard_by_instance(instance: ShardMixin) -> str:
    key = mod_shard_key_by_replica_count(shard_key=get_shard_key_from_instance(instance=instance), shard_group=instance.shard_group)
    databases = get_master_databases_by_shard_group(shard_group=instance.shard_group)
    return databases[_calc_shard(shard_key=key, shard_count=len(databases))]


def _calc_shard(shard_key: int, shard_count: int) -> int:
    return int(shard_key % shard_count)



from typing import Optional

from django.conf import settings

from shard.strategy.routing.random import RandomReadStrategy


class BaseRouter:
    def db_for_read(self, model, **hints):
        raise NotImplementedError

    def db_for_write(self, model, **hints):
        raise NotImplementedError

    def allow_relation(self, obj1, obj2, **hints):
        raise NotImplementedError

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        raise NotImplementedError


class BaseReplicationRouter(BaseRouter):
    _strategy = RandomReadStrategy

    def db_for_read(self, model, **hints):
        master = self._get_master_database(model=model, **hints)

        if master is None:
            return None

        return self._get_slave_database(master=master)

    def db_for_write(self, model, **hints):
        return self._get_master_database(model=model, **hints)

    def allow_relation(self, obj1, obj2, **hints):
        obj1_db = self._get_master_database(model=obj1.__class__, instance=obj1)
        obj2_db = self._get_master_database(model=obj2.__class__, instance=obj2)

        # If variable is none, has not opinion for allow_relation.
        if obj1_db is None or obj2_db is None:
            return None

        return obj1_db == obj2_db

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        # If db is slave, return false
        if settings.DATABASES[db].get('PRIMARY', None):
            return False

        return None

    def _get_master_database(self, model, **hints) -> Optional[str]:
        raise NotImplementedError

    def _get_slave_database(self, master: str) -> str:
        return self._strategy.pick_read_db(master=master)



import uuid

from django.utils.deconstruct import deconstructible

from shard.strategy.id_generation.base import BaseIDGenerationStrategy

__all__ = ('UUIDStrategy', )


@deconstructible
class UUIDStrategy(BaseIDGenerationStrategy):
    def get_next_id(self, instance):
        return uuid.uuid4().hex



from setuptools import setup, find_packages

version = '0.0.8'

with open('requirements/base.txt') as f:
    install_requires = [line for line in f if line and not line.startswith('-')]


setup(
    name='django-shard-library',
    version=version,
    url='https://github.com/ridi/django-shard-library',
    license='BSD',
    description='Django Shard Library',
    keywords=['django', 'shard', 'ridi', 'ridibooks'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',

        'License :: OSI Approved :: BSD License',
        'Framework :: Django :: 2.0',
        'Programming Language :: Python :: 3.6',
    ],
    packages=find_packages(exclude=('tests', 'docs',)),
    install_requires=install_requires,
    include_package_data=True,
)



from django.db import models

from shard.exceptions import StrategyNotImplementException, RequireSpecificDatabaseException

__all__ = ('StrategyPkMixin', 'SpecificDatabaseMixin', 'ShardMixin', )


class StrategyPkMixin(models.Field):
    def __init__(self, *args, **kwargs):
        strategy = kwargs.get('strategy', None)
        if strategy is None:
            raise StrategyNotImplementException

        self.strategy = strategy
        del kwargs['strategy']
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        kwargs['strategy'] = self.strategy
        return name, path, args, kwargs

    def get_pk_value_on_save(self, instance):
        if not instance.pk:
            return self.strategy.get_next_id(instance=instance)
        return instance.pk


class SpecificDatabaseMixin:
    _specific_database = None

    @classmethod
    def specific_database(cls):
        return cls._get_database()

    @classmethod
    def _get_database(cls):
        if cls._specific_database is None:
            raise RequireSpecificDatabaseException

        return cls._specific_database


class ShardMixin:
    shard_group = None
    shard_key_name = None

    def get_shard_key(self):
        return getattr(self, self.shard_key_name)



import random

from shard.strategy.routing.base import BaseReadRoutingStrategy


class RandomReadStrategy(BaseReadRoutingStrategy):
    @classmethod
    def pick_read_db(cls, master: str):
        databases = cls.get_slave_databases(master=master)

        if not databases:
            return master

        return random.choice(databases)



from uuid import uuid4

from shard_static.constants import DEFAULT_LOCK_TTL


class BaseLockManager:
    def __init__(self, key: str, ttl: int=DEFAULT_LOCK_TTL):
        self.key = key
        self.uuid = uuid4().hex
        self.ttl = ttl

    def is_locked(self) -> bool:
        raise NotImplementedError()

    def lock(self) -> bool:
        raise NotImplementedError()

    def ping(self) -> bool:
        raise NotImplementedError()

    def release(self) -> bool:
        raise NotImplementedError()



import random
import time

from django.utils.deconstruct import deconstructible
from shard.utils.shard_key import get_shard_key_from_instance

from shard.strategy.id_generation.base import BaseIDGenerationStrategy

__all__ = ('NanosecondStrategy', )


_NANO_SECOND = 1000000
_SHARD_KEY_BIT_LEN = (1 << 10)

_NANO_SECOND_PLACE_VALUE = 13
_SHARD_KEY_BIT_PLACE_VALUE = 3


@deconstructible
class NanosecondStrategy(BaseIDGenerationStrategy):
    def get_next_id(self, instance):
        # nanosecond 51 bit
        # shard_key_bit 10 bit
        # randint 3 bit
        # total 64 bit

        nanosecond = int(time.time() * _NANO_SECOND)
        shard_key_bit = get_shard_key_from_instance(instance) % _SHARD_KEY_BIT_LEN

        _rand = random.Random(id(object()))
        _randint = _rand.randint(0, 7)

        return nanosecond << _NANO_SECOND_PLACE_VALUE | shard_key_bit << _SHARD_KEY_BIT_PLACE_VALUE | _randint



from django.db.models.query import QuerySet

__all__ = ('ShardQuerySet', )


class ShardQuerySet(QuerySet):
    def __init__(self, model, *args, **kwargs):
        super().__init__(model=model, *args, **kwargs)
        self._instance = None

    @property
    def db(self):
        if self._db:
            return self._db

        if not self._hints.get('instance') and getattr(self, '_instance', None):
            self._hints['instance'] = getattr(self, '_instance')

        return super().db

    def create(self, **kwargs):
        self._instance = self.model(**kwargs)
        return super().create(**kwargs)



from django.db import models
from django.db.models import BigAutoField

from shard.constants import DEFAULT_DATABASE
from shard.mixins import SpecificDatabaseMixin


class TableStrategyModel(SpecificDatabaseMixin, models.Model):
    id = BigAutoField(primary_key=True)

    _specific_database = DEFAULT_DATABASE

    class Meta:
        abstract = True
